#include <iostream.h>
#include <pthread.h>
#include <string.h>
#include <unistd.h>

#include <FinalSolution.h>
#include <Debug.h>

static FinalSolution *FS;


pwdnumdiv_t pwdnumdiv(pwdnum num, pwdnum den)
{
	static pwdnumdiv_t d;
	pwdnum i;

	d.quot = d.rem = 0;
	if(den == 0 && num == 0)
	{
		return d;
	}

	d.quot = num / den;
	d.rem = num % den;

	return d;
}


FinalSolution::FinalSolution()
{
	started = initialized = hostchecked = nomorepasswds = passwordfound = false;
	threads = free_threads = 0;
	h = NULL;
	pthread_mutex_init(&mutex, NULL);
}


FinalSolution::~FinalSolution()
{
}


void FinalSolution::SetupCrack(
	Protocol myproto,
	ProtocolParams *myparams,
	PasswdGen *mypwg,
	int mythreads)
{
	if(started) return;
	
	proto = myproto;
	params = myparams;
	pwg = mypwg;
	free_threads = threads = mythreads;

	initialized = true;
}


void FinalSolution::StartCrack()
{
	if(!initialized) return;
	if(started) return;

	textiface.ExecInterface();
	started = true;
	nomorepasswds = false;

	// create the first threads
	for(int i=0; i<threads; i++)
	{
		pthread_t this_thread;
		pthread_attr_t attr;

		pthread_attr_init(&attr);
		pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

		if(
		pthread_create(
			&this_thread,
			&attr,
			CheckerThread,
			NULL) == -1)
		{
			debug("Thread creation error!! -Initial threads creation-")
		}
	}
	
	// create a new thread when one thread finishes, until the password is
	// found or nomorepasswds==true
	for(;;)
	{
		if(nomorepasswds) break;
		if(passwordfound) break;

		if(free_threads>0)
		{
			pthread_t this_thread;
			pthread_attr_t attr;

			pthread_attr_init(&attr);
			pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

			if(
			pthread_create(
				&this_thread,
				&attr,
				CheckerThread,
				NULL) == -1)
			{
				debug("Thread creation error!!")
			}

			//free_threads--;
		}

	}

	// wait all the threads to finish
	for(;;)
	{
		//cout << "Free: " << free_threads << " of " << threads << endl;
		if(free_threads == threads) break;
	}

	/*if(nomorepasswds)*/ textiface.CrackFinished();
	started = false;
}


void FinalSolution::PauseCrack()
{
}


char *FinalSolution::GetPassword()
{
	if(passwordfound) return password;
	else return NULL;
}


// Private functions


void FinalSolution::ThreadFinished(
	ConnectionResult result,
	char mypassword[BIGGEST_PASSWORD])
{
	pthread_mutex_lock(&mutex);
	
//	free_threads++;
//	debug("Thread finished. free_threads=" << free_threads)

	switch(result)
	{
		case Login:
		passwordfound = true;
		strncpy(password, mypassword, BIGGEST_PASSWORD);
		textiface.FoundPasswd();
		break;

		case PasswordIncorrect:
		if(!passwordfound) textiface.OnePasswdLess(mypassword);
		break;

		case NoMorePasswds:
		if(!nomorepasswds)
		{
			nomorepasswds = true;
			//textiface.CrackFinished();
		}
		break;

		case UserIncorrect:
		textiface.UserIncorrect();
		break;

		case HostNotFound:
		textiface.HostNotFound();
		break;

		case ConnectionRefused:
		textiface.ConnectionRefused();
		break;

		case TimeOut:
		textiface.TimeOut();
		break;
	}

	free_threads++;
	debug("Thread finished. free_threads=" << free_threads)

	pthread_mutex_unlock(&mutex);
	pthread_exit(0);
}


void *CheckerThread(void *ptr)
{
	ProtocolChecker *checker;
	char password[BIGGEST_PASSWORD];
	bool nomorepasswds;

	FS->free_threads--;
	debug("Thread started. PID: " << pthread_self() << ". free_threads=" << FS->free_threads)

	if(FS->nomorepasswds) FS->ThreadFinished(NoMorePasswds, password);


	switch(FS->proto)
	{
		case FTP:
		checker = new FTPProtocolChecker;
		break;
		
		case POP3:
		break;

		default:
		break;
	}

	FS->pwg->NextPasswd(password, &nomorepasswds);
	
	if(nomorepasswds == true)
	{
		debug("No more passwords")
		FS->ThreadFinished(NoMorePasswds, password);
	}
	
	checker->Setup(FS->params);
	int i;

	ConnectionResult result;

	for(i=0; i<MAX_ATTEMPTS; i++)
	{
		result = checker->Try(password);

		if(result==HostNotFound)
		{
			FS->textiface.HostNotFound();
			sleep(RETRY_DELAY);
		} else
		if(result==ConnectionRefused)
		{
			FS->textiface.ConnectionRefused();
			sleep(RETRY_DELAY);
		} else break;
	}

	if(result==HostNotFound)
	{
		FS->textiface.HostNotFound();
		FS->textiface.CouldNotConnect();
		exit(0);
	} 
	if(result==ConnectionRefused)
	{
		FS->textiface.ConnectionRefused();
		FS->textiface.CouldNotConnect();
		exit(0);
	}

	FS->ThreadFinished(result, password);
}


// Non-class functions


void SetFinalSolution(FinalSolution *myfs)
{
	FS=myfs;
}


FinalSolution *GetFinalSolution()
{
	return FS;
}

